@@ -41,7 +41,7 @@ static void stat_override_copy( const ptlib_stat *stat, stat_override *override
     override->dev=stat->dev;
     override->inode=stat->ino;
     override->uid=stat->uid;
-    override->gid=stat->uid;
+    override->gid=stat->gid;
     override->dev_id=stat->rdev;
     override->mode=stat->mode;
 }
@@ -53,7 +53,7 @@ bool sys_stat( int sc_num, pid_t pid, pid_state *state )
         // Entering the syscall
         state->state=pid_state::RETURN;
         state->context_state[0]=ptlib_get_argument( pid, 2 ); // Store the pointer to the stat struct
-        dlog("stat64: "PID_F" stored pointer at %p\n", pid, state->context_state[0] );
+        dlog("stat64: "PID_F" stored pointer at %p\n", pid, (void*)state->context_state[0] );
 
         // If the process is chrooted, we need to translate the file name
         int real_sc=ptlib_get_syscall( pid );
@@ -84,10 +84,16 @@ bool sys_stat( int sc_num, pid_t pid, pid_state *state )
                     // If the override is not a device, and the types do not match, this is not a valid entry
                     ok=(S_IFMT&ret.mode)==(S_IFMT&override.mode);
                 }
-                ret.mode=(ret.mode&(~(07000|S_IFMT))) | (override.mode&(07000|S_IFMT));
+                // Override the u=x flag for directories, but not files
+                if( S_ISDIR(ret.mode) ) {
+                    ret.mode=(ret.mode&(~(07700|S_IFMT))) | (override.mode&(07700|S_IFMT));
+                } else {
+                    ret.mode=(ret.mode&(~(07600|S_IFMT))) | (override.mode&(07600|S_IFMT));
+                }
 
                 if( ok ) {
-                    dlog("stat64: "PID_F" override dev="DEV_F" inode="INODE_F" mode=%o uid=%d gid=%d\n", pid, ret.dev, ret.ino, ret.mode, ret.uid, ret.gid );
+                    dlog("stat64: "PID_F" override dev="DEV_F" inode="INODE_F" mode=%o uid="UID_F" gid="GID_F"\n",
+                        pid, ret.dev, ret.ino, ret.mode, ret.uid, ret.gid );
                     ptlib_set_mem( pid, &ret, (void *)state->context_state[0], sizeof(struct stat) );
                 } else {
                     dlog("stat64: "PID_F" dev="DEV_F" inode="INODE_F" override entry corrupt - removed\n", pid, ret.dev, ret.ino );
@@ -111,7 +117,7 @@ bool sys_fstatat64( int sc_num, pid_t pid, pid_state *state )
         chroot_translate_paramat( pid, state, ptlib_get_argument( pid, 1 ), 2, (ptlib_get_argument(pid, 4)&AT_SYMLINK_NOFOLLOW)!=0 );
 
         state->context_state[0]=ptlib_get_argument( pid, 3 ); // Store the pointer to the stat struct
-        dlog("statat64: "PID_F" stored pointer at %p\n", pid, state->context_state[0] );
+        dlog("statat64: "PID_F" stored pointer at %p\n", pid, (void*)state->context_state[0] );
 
         return true;
     } else {
@@ -127,57 +133,105 @@ static bool real_chmod( int sc_num, pid_t pid, pid_state *state, int mode_offset
             return allocate_process_mem( pid, state, sc_num );
         }
 
-        mode_t mode=(mode_t)ptlib_get_argument( pid, mode_offset+1 ); // Store the requested mode
-        state->context_state[0]=mode;
+        // First we stat the file to find out what we are up against (dev/inode etc.)
+        state->state=pid_state::REDIRECT2;
+        state->orig_sc=sc_num;
 
-        mode=mode&~07000;
-        ptlib_set_argument( pid, mode_offset+1, mode ); // Zero out the S* field
+        state->context_state[1]=ptlib_get_argument( pid, mode_offset+1 ); // Store the requested mode
+        state->context_state[0]=0; // syscall parts progress
 
-        dlog("chmod: "PID_F" mode %o changed to %o\n", pid, state->context_state[1], mode );
-        state->state=pid_state::RETURN;
-    } else if( state->state==pid_state::RETURN ) {
+        ptlib_set_argument( pid, mode_offset+1, (int_ptr)state->memory ); // where to store the stat result
+
+        ptlib_set_syscall( pid, stat_function );
+
+    } else if( state->state==pid_state::REDIRECT2 && state->context_state[0]==0 ) {
         if( ptlib_success( pid, sc_num ) ) {
-            dlog("chmod: "PID_F" chmod successful, performing stat so we can update the map\n", pid);
+            // Our stat succeeded
+            struct stat_override override;
+            struct ptlib_stat stat;
+
+            ptlib_get_mem( pid, state->memory, &stat, sizeof( stat ) );
 
-            // We need to call "stat/fstat" so we can know the dev/inode
-            state->state=pid_state::REDIRECT1;
             ptlib_save_state( pid, state->saved_state );
-            state->orig_sc=sc_num;
 
+            // Restore the original parameters to restart the actual chmod call
             for( int i=1; i<=mode_offset; ++i )
-                ptlib_set_argument( pid, i, state->context_state[i] );
+                ptlib_set_argument( pid, i, state->context_state[i+1] );
 
-            ptlib_set_argument( pid, mode_offset+1, (int_ptr)state->memory ); // Where to store the stat result
-
-            // One anomaly handled with special case. Ugly, but not worth the interface complication
+            // one anomaly handled with special case. ugly, but not worth the interface complication
             if( extra_flags!=-1 ) {
-                // Some of the functions require an extra flag after the usual parameters
+                // some of the functions require an extra flag after the usual parameters
                 ptlib_set_argument( pid, mode_offset+2, extra_flags );
             }
 
-            return ptlib_generate_syscall( pid, stat_function, state->shared_memory );
+            state->state=pid_state::REDIRECT1;
+
+            // Modify the chmod mode
+            mode_t mode=(mode_t)state->context_state[1];
+
+            mode&=~07000; // Clear the SUID etc. bits
+            if( S_ISDIR( stat.mode ) ) {
+                // The node in question is a directory
+                mode|=00700; // Make sure we have read, write and execute permission
+            } else {
+                // Node is not a directory
+                mode|=00600; // Set read and write for owner
+            }
+            ptlib_set_argument( pid, mode_offset+1, mode );
+
+            // Save the stuff we'll need in order to update the lies database
+            state->context_state[2]=stat.dev;
+            state->context_state[3]=stat.ino;
+
+            state->context_state[0]=1; // Mark this as the actual chmod call
+
+            // If we don't already have an entry for this file in the lies database, we will not have
+            // the complete stat struct later on to create it.
+            if( !get_map( stat.dev, stat.ino, &override ) ) {
+                // Create a lie that is identical to the actual file
+                stat_override_copy( &stat, &override );
+                set_map( &override );
+            }
+
+            return ptlib_generate_syscall( pid, state->orig_sc, state->shared_memory );
         } else {
+            // stat failed - return that failure to the caller as is
             state->state=pid_state::NONE;
-            dlog("chmod: "PID_F" chmod failed with error %s\n", pid, strerror(ptlib_get_error(pid, sc_num)));
         }
-    } else if( state->state==pid_state::REDIRECT2 ) {
-        // Update our lies database
-        struct stat_override override;
-        struct ptlib_stat stat;
+    } else if( state->state==pid_state::REDIRECT2 && state->context_state[0]==1 ) {
+        if( ptlib_success( pid, sc_num ) ) {
+            // The chmod call succeeded - update the lies database
+            struct stat_override override;
 
-        ptlib_get_mem( pid, state->memory, &stat, sizeof( stat ) );
+            if( !get_map( state->context_state[2], state->context_state[3], &override ) ) {
+                // We explicitly created this map not so long ago - something is wrong
+                // XXX What can we do except hope these are reasonable values
+                override.dev=state->context_state[2];
+                override.inode=state->context_state[3];
+                override.uid=0;
+                override.gid=0;
+                override.dev_id=0;
+                override.transient=true;
 
-        if( !get_map( stat.dev, stat.ino, &override ) ) {
-            stat_override_copy( &stat, &override );
+                dlog("chmod: "PID_F" error (race?) getting override info for dev "DEV_F" inode "INODE_F"\n",
+                        pid, state->context_state[2], state->context_state[3] );
+            }
+            override.mode=(override.mode&~07777)|(((mode_t)state->context_state[1])&07777);
+            dlog("chmod: "PID_F" Setting override mode %o dev "DEV_F" inode "INODE_F"\n", pid, override.mode, override.dev,
+                    override.inode );
+            set_map( &override );
+
+            ptlib_restore_state( pid, state->saved_state );
+            ptlib_set_retval( pid, 0 );
+        } else {
+            // We just need to restore the saved state but keep the chmod error code
+            int err=ptlib_get_error( pid, sc_num );
+
+            ptlib_restore_state( pid, state->saved_state );
+            ptlib_set_error( pid, sc_num, err );
         }
-        override.mode=(override.mode&~07777)|(((mode_t)state->context_state[0])&07777);
-
-        dlog("chmod: "PID_F" Setting override mode %o dev "DEV_F" inode "INODE_F"\n", pid, override.mode, override.dev,
-            override.inode );
-        set_map( &override );
 
         state->state=pid_state::NONE;
-        ptlib_restore_state( pid, state->saved_state );
     } else {
         dlog("chmod: "PID_F" unknown state %d\n", pid, state->state );
     }
@@ -191,7 +245,7 @@ bool sys_chmod( int sc_num, pid_t pid, pid_state *state )
     if( state->state==pid_state::NONE ) {
         chroot_translate_param( pid, state, 1, true );
 
-        state->context_state[1]=ptlib_get_argument( pid, 1 ); // Store the file name
+        state->context_state[2]=ptlib_get_argument( pid, 1 ); // Store the file name
     }
 
     return real_chmod( sc_num, pid, state, 1, PREF_STAT );
@@ -200,7 +254,7 @@ bool sys_chmod( int sc_num, pid_t pid, pid_state *state )
 bool sys_fchmod( int sc_num, pid_t pid, pid_state *state )
 {
     if( state->state==pid_state::NONE ) {
-        state->context_state[1]=ptlib_get_argument( pid, 1 ); // Store the file descriptor
+        state->context_state[2]=ptlib_get_argument( pid, 1 ); // Store the file descriptor
     }
 
     return real_chmod( sc_num, pid, state, 1, PREF_FSTAT );
@@ -214,12 +268,12 @@ bool sys_fchmodat( int sc_num, pid_t pid, pid_state *state )
         // this chroot translation will need to be reconsidered
         chroot_translate_paramat( pid, state, ptlib_get_argument( pid, 1), 2, true ); 
 
-        state->context_state[1]=ptlib_get_argument( pid, 1 ); // Store the base dir fd
-        state->context_state[2]=ptlib_get_argument( pid, 2 ); // Store the file name
-        state->context_state[3]=ptlib_get_argument( pid, 4 ); // Store the flags
+        state->context_state[2]=ptlib_get_argument( pid, 1 ); // Store the base dir fd
+        state->context_state[3]=ptlib_get_argument( pid, 2 ); // Store the file name
+        state->context_state[4]=ptlib_get_argument( pid, 4 ); // Store the flags
     }
 
-    return real_chmod( sc_num, pid, state, 2, PREF_FSTATAT, state->context_state[3] );
+    return real_chmod( sc_num, pid, state, 2, PREF_FSTATAT, 0 );
 }
 #endif
 
@@ -332,17 +386,17 @@ static bool real_mknod( int sc_num, pid_t pid, pid_state *state, int mode_offset
 
         mode_t mode=(mode_t)state->context_state[0];
 
-        if( (mode&07000)!=0 ) {
-            // Mode has a SUID set
-            mode&=~(07000);
-        }
+        // Remove SUID and add user read and write
+        mode&=~07000;
+        mode|= 00600;
+
         if( S_ISCHR(mode) || S_ISBLK(mode) ) {
             dlog("mknod: "PID_F" tried to create %s device, turn to regular file\n", pid, S_ISCHR(mode) ? "character" : "block" );
             mode=(mode&~S_IFMT) | S_IFREG;
         }
         ptlib_set_argument( pid, mode_offset+1, mode );
 
-        dlog("mknod: %d mode %o\n", pid, state->context_state[1] );
+        dlog("mknod: %d mode %o\n", pid, (unsigned int)state->context_state[1] );
         state->state=pid_state::RETURN;
     } else if( state->state==pid_state::RETURN ) {
         if( ptlib_success( pid, sc_num ) ) {
@@ -389,6 +443,10 @@ static bool real_mknod( int sc_num, pid_t pid, pid_state *state, int mode_offset
                 override.mode=(override.mode&~(S_IFMT|07000)) | (mode&(S_IFMT|07000));
                 override.dev_id=(dev_t)state->context_state[1];
             }
+            // use the user read+write from the original, not the actual file
+            // XXX This code disregards the umask
+            override.mode&=~00600;
+            override.mode|= mode&00600;
 
             set_map( &override );
         } else {
@@ -432,9 +490,23 @@ bool sys_mknodat( int sc_num, pid_t pid, pid_state *state )
 }
 #endif
 
-static bool real_open( int sc_num, pid_t pid, pid_state *state )
+static bool real_open( int sc_num, pid_t pid, pid_state *state, int mode_argnum )
 {
     if( state->state==pid_state::NONE ) {
+        if( (state->context_state[0]&O_CREAT)!=0 ) {
+            // We are asking to create the file - make sure we give user read and write permission
+            state->context_state[1]=ptlib_get_argument( pid, mode_argnum );
+            mode_t mode=state->context_state[1];
+            
+            // Remove SUID and add user read and write permission
+            mode&=~07000;
+            mode|= 00600;
+
+            if( mode!=state->context_state[1] ) {
+                ptlib_set_argument( pid, mode_argnum, mode );
+            }
+        }
+
         state->state=pid_state::RETURN;
     } else if( state->state==pid_state::RETURN ) {
         // Did we request to create a new file?
@@ -468,6 +540,9 @@ static bool real_open( int sc_num, pid_t pid, pid_state *state )
 
                 override.uid=0;
                 override.gid=0;
+                override.mode&=~07600;
+                override.mode|= state->context_state[1]&07600;
+                // XXX We are ignoring the umask here!
 
                 set_map( &override );
                 dlog("open: "PID_F" creating override for dev "DEV_F" inode "INODE_F"\n", pid, override.dev, override.inode);
@@ -493,7 +568,7 @@ bool sys_open( int sc_num, pid_t pid, pid_state *state )
         state->context_state[0]=ptlib_get_argument( pid, 2 ); //flags
     }
 
-    return real_open( sc_num, pid, state );
+    return real_open( sc_num, pid, state, 3 );
 }
 
 #if HAVE_OPENAT
@@ -505,7 +580,7 @@ bool sys_openat( int sc_num, pid_t pid, pid_state *state )
         state->context_state[0]=ptlib_get_argument( pid, 3 ); //flags
     }
 
-    return real_open( sc_num, pid, state );
+    return real_open( sc_num, pid, state, 4 );
 }
 #endif
 
@@ -516,6 +591,17 @@ static bool real_mkdir( int sc_num, pid_t pid, pid_state *state, int mode_offset
         if( state->memory==NULL )
             return allocate_process_mem( pid, state, sc_num );
 
+        // Make sure user has rwx on the created directory
+        state->context_state[0]=ptlib_get_argument( pid, mode_offset+1 );
+
+        mode_t mode=state->context_state[0];
+        mode|= 00700;
+
+        if( mode!=state->context_state[0] ) {
+            // We did change the mode
+            ptlib_set_argument( pid, mode_offset+1, mode );
+        }
+
         state->state=pid_state::RETURN;
     } else if( state->state==pid_state::RETURN ) {
         state->state=pid_state::NONE;
@@ -525,8 +611,8 @@ static bool real_mkdir( int sc_num, pid_t pid, pid_state *state, int mode_offset
             ptlib_save_state( pid, state->saved_state );
 
             // Perform a stat operation so we can know the directory's dev and inode
-            for( int i=0; i<mode_offset; ++i )
-                ptlib_set_argument( pid, i+1, state->context_state[i] ); // Name
+            for( int i=1; i<=mode_offset; ++i )
+                ptlib_set_argument( pid, i, state->context_state[i] ); // The original mkdir arguments
             ptlib_set_argument( pid, mode_offset+1, (int_ptr)state->memory ); // stat structure
 
             if( extra_flags!=-1 ) {
@@ -553,6 +639,10 @@ static bool real_mkdir( int sc_num, pid_t pid, pid_state *state, int mode_offset
             override.uid=0;
             override.gid=0;
 
+            override.mode&=~00700;
+            override.mode|= state->context_state[0]&00700;
+            // XXX This code does not take the umask into account
+
             dlog("mkdir: "PID_F" storing override for dev "DEV_F" inode "INODE_F"\n", pid, override.dev, override.inode);
             set_map( &override );
         } else {
@@ -571,7 +661,7 @@ bool sys_mkdir( int sc_num, pid_t pid, pid_state *state )
     if( state->state==pid_state::NONE && state->memory!=NULL ) {
         chroot_translate_param( pid, state, 1, true );
 
-        state->context_state[0]=ptlib_get_argument( pid, 1 ); // Directory name
+        state->context_state[1]=ptlib_get_argument( pid, 1 ); // Directory name
 
         if( log_level>0  ) {
             char name[PATH_MAX];
@@ -591,13 +681,14 @@ bool sys_mkdirat( int sc_num, pid_t pid, pid_state *state )
     if( state->state==pid_state::NONE ) {
         chroot_translate_paramat( pid, state, ptlib_get_argument( pid, 1 ), 2, true );
 
-        state->context_state[0]=ptlib_get_argument( pid, 2 ); // Directory name
+        state->context_state[1]=ptlib_get_argument( pid, 1 ); // File descriptor
+        state->context_state[2]=ptlib_get_argument( pid, 2 ); // Directory name
 
         if( log_level>0  ) {
             char name[PATH_MAX];
 
-            ptlib_get_string( pid, (void *)state->context_state[0], name, sizeof(name) );
-            int fd=(int)ptlib_get_argument( pid, 1 );
+            ptlib_get_string( pid, (void *)state->context_state[2], name, sizeof(name) );
+            int fd=(int)state->context_state[1];
 
             dlog("mkdirat: %d creates %s at %x\n", pid, name, fd );
         }
@@ -752,6 +843,7 @@ bool sys_getcwd( int sc_num, pid_t pid, pid_state *state )
 
 bool sys_munmap( int sc_num, pid_t pid, pid_state *state )
 {
+    // XXX Does this work? We should never be called with REDIRECT1
     if( state->state==pid_state::NONE || state->state==pid_state::REDIRECT1 ) {
         if( state->state==pid_state::NONE ) {
             // This is our first time entering this munmap
@@ -794,7 +886,7 @@ bool sys_munmap( int sc_num, pid_t pid, pid_state *state )
         if( start<addr1 && end>addr1 ) {
             // The unmap range covers the lower range
             dlog("sys_munmap: "PID_F" tried to unmap range %p-%p, which conflicts with our %s range %p-%p\n",
-                pid, start, end, name1, addr1, addr1+len1 );
+                pid, (void*)start, (void*)end, name1, (void*)addr1, (void*)(addr1+len1) );
 
             if( end>addr1+len1 ) {
                 // There is an area to unmap above the lower range - mark it for a second syscall
@@ -806,13 +898,13 @@ bool sys_munmap( int sc_num, pid_t pid, pid_state *state )
         } else if( start>=addr1 && start<addr1+len1 ) {
             // The start pointer is inside the lower memory range
             dlog("sys_munmap: "PID_F" tried to unmap range %p-%p, which conflicts with our %s range %p-%p\n",
-                pid, start, end, name1, addr1, addr1+len1 );
+                pid, (void*)start, (void*)end, name1, (void*)addr1, (void*)(addr1+len1) );
 
             start=addr1+len1;
         } else if( start<addr2 && end>addr2 ) {
             // The unmap area covers the upper memory range
             dlog("sys_munmap: "PID_F" tried to unmap range %p-%p, which conflicts with our %s range %p-%p\n",
-                pid, start, end, name2, addr2, addr2+len2 );
+                pid, (void*)start, (void*)end, name2, (void*)addr2, (void*)(addr2+len2) );
 
             if( end>addr2+len2 ) {
                 // There is an area to unmap above the upper range - mark it for a second (third?) syscall
@@ -824,7 +916,7 @@ bool sys_munmap( int sc_num, pid_t pid, pid_state *state )
         } else if( start>=addr2 && start<addr2+len2 ) {
             // The start pointer is inside the upper memory range
             dlog("sys_munmap: "PID_F" tried to unmap range %p-%p, which conflicts with our %s range %p-%p\n",
-                pid, start, end, name2, addr2, addr2+len2 );
+                pid, (void*)start, (void*)end, name2, (void*)addr2, (void*)(addr2+len2) );
 
             start=addr2+len2;
         } else {
@@ -1109,6 +1201,14 @@ bool sys_unlinkat( int sc_num, pid_t pid, pid_state *state )
                 if( ptlib_success( pid, sc_num ) ) {
                     ptlib_stat stat;
 
+                    // We are tight on context memory - first set up the args for the continuation syscall,
+                    // only then handle the current call return
+                    ptlib_save_state( pid, state->saved_state );
+                    ptlib_set_argument( pid, 1, state->context_state[2] );
+                    ptlib_set_argument( pid, 2, state->context_state[3] );
+                    ptlib_set_argument( pid, 3, state->context_state[0]==10 ? AT_REMOVEDIR : 0 );
+
+                    // Will the file be actually deleted?
                     ptlib_get_mem( pid, state->memory, &stat, sizeof(stat) );
 
                     if( stat.nlink==1 || state->context_state[0]==10 ) {
@@ -1122,11 +1222,7 @@ bool sys_unlinkat( int sc_num, pid_t pid, pid_state *state )
                         state->context_state[2]=0;
                     }
 
-                    // Perform the actual unlink operation
-                    ptlib_save_state( pid, state->saved_state );
-                    ptlib_set_argument( pid, 1, state->context_state[2] );
-                    ptlib_set_argument( pid, 2, state->context_state[3] );
-                    ptlib_set_argument( pid, 3, state->context_state[0]==10 ? AT_REMOVEDIR : 0 );
+                    // Complete the continuation call
                     ptlib_generate_syscall( pid, state->orig_sc, state->shared_memory );
                     state->state=pid_state::REDIRECT1;
                     state->context_state[0]=1;
