@@ -27,8 +27,10 @@
 #include <sys/syscall.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
+#include <sys/socket.h>
 
 #include MAP_INCLUDE
+#include <set>
 
 #include <stdio.h>
 #include <assert.h>
@@ -53,10 +55,28 @@ static bool handle_memory_allocation( int sc_num, pid_t pid, pid_state *state );
 static MAP_CLASS<int, syscall_hook> syscalls;
 
 // Keep track of the states for the various processes
-static MAP_CLASS<pid_t, pid_state> state;
+template <class key, class data> class map_class : public  MAP_CLASS<key, data> 
+{
+    // Inherit everything, just disable the dangerous operator[]
+public:
+    data &operator[] (const key &k)
+    {
+        return MAP_CLASS<key,data>::operator[] (k);
+    }
+    const data &operator[] ( const key &k) const
+    {
+        return MAP_CLASS<key,data>::operator[] (k);
+    }
+};
+
+static map_class<pid_t, pid_state> state;
 
 size_t static_mem_size, shared_mem_size;
 
+static MAP_CLASS<pid_t, int> root_children; // Map of all root children
+
+static int num_processes; // Number of running processes
+
 static void init_handlers()
 {
     syscalls[SYS_geteuid]=syscall_hook(sys_getuid, "geteuid");
@@ -280,6 +300,7 @@ static const char *state2str( pid_state::states state )
         STATENAME(ALLOCATE)
         STATENAME(ALLOC_RETURN)
         STATENAME(WAITING)
+        STATENAME(ZOMBIE)
 #undef STATENAME
     }
 
@@ -308,7 +329,9 @@ static void notify_parent( pid_t parent, const pid_state::wait_state &waiting )
         return;
     }
     dlog("notify_parent: "PID_F" sent a notify about "PID_F"(%x)\n", parent, waiting.pid(), waiting.status());
-    pid_state *proc_state=&state[parent];
+    pid_state *proc_state=lookup_state(parent);
+    assert(proc_state!=NULL);
+
     proc_state->waiting_signals.push_back( waiting );
 
     // Is the parent currently waiting?
@@ -328,7 +351,14 @@ static void handle_exit( pid_t pid, int status, const struct rusage &usage )
     dlog(NULL);
     assert(proc_state!=NULL);
 
-    // First thing first - notify the parent
+    // Set the process state to ZOMBIE with usage count of 1
+    proc_state->state=pid_state::ZOMBIE;
+    proc_state->context_state[0]=1;
+    dlog("%s: "PID_F" is now a zombie\n", __func__, pid );
+
+    pid_state *parent_state=lookup_state(proc_state->parent);
+
+    // Notify the parent
 #if PTLIB_PARENT_CAN_WAIT
     // If a parent can wait on a debugged child we need to notify it even if the child is being debugged,
     // but only if it actually has a parent (i.e. - was not reparented to init)
@@ -338,98 +368,180 @@ static void handle_exit( pid_t pid, int status, const struct rusage &usage )
     // If a parent cannot wait, we need to let it know ourselves only if it's not being debugged
     if( (proc_state->debugger==0 || proc_state->debugger==proc_state->parent) && proc_state->parent!=0 && proc_state->parent!=1 )
 #endif
+    {
+        proc_state->context_state[0]++; // Update use count
         notify_parent( proc_state->parent, pid_state::wait_state( pid, status, &usage, false ) );
+    }
+
+    // Regardless of whether it is being notified or not, the parent's child num needs to be decreased
+    if( parent_state!=NULL ) {
+        parent_state->num_children--;
+    }
 
     if( proc_state->debugger!=0 && proc_state->debugger!=proc_state->parent ) {
         // The process was being debugged - notify the debugger as well
+        proc_state->context_state[0]++; // Update use count
         notify_parent( proc_state->parent, pid_state::wait_state( pid, status, &usage, true ) );
         state[proc_state->debugger].num_debugees--;
     }
 
-    pid_state *parent_state;
-    // Regardless of whether it is being notified or not, the parent's child num needs to be decreased
-    if( proc_state->parent!=0 && proc_state->parent!=1 && (parent_state=lookup_state(proc_state->parent))!=NULL ) {
-        parent_state->num_children--;
-    }
-
     // Is any process a child of this process?
+    // We need to delete all child zombie processes. This means changing the list while scanning it.
+    // Instead, create a list of pids to delete
+    std::set<pid_t> need_delete;
     for( MAP_CLASS<pid_t, pid_state>::iterator i=state.begin(); i!=state.end(); ++i ) {
         if( i->second.parent==pid ) {
-            dlog("Reparenting process %d to init from %d\n", i->first);
+            dlog("Reparenting process %d to init from %d\n", i->first, pid);
             i->second.parent=1;
-        }
 
+            if( i->second.state==pid_state::ZOMBIE ) {
+                // "init" should release it
+                need_delete.insert(i->first);
+            }
+        } 
+        
         if( i->second.debugger==pid ) {
             dlog("Detaching process %d from recursive debugger %d\n", i->first, pid );
             i->second.debugger=0;
+
+            if( i->second.state==pid_state::ZOMBIE && i->second.parent!=pid ) {
+                // The process is in zombie state, pid is its debugger but not parent
+                need_delete.insert(i->first);
+            }
         }
     }
-    state.erase(pid);
+
+    for( std::set<pid_t>::iterator i=need_delete.begin(); i!=need_delete.end(); ++i ) {
+        delete_state(*i);
+    }
+
+    // Delete the state from our end. The state is reference counted, so it may not actually be deleted just yet
+    delete_state(pid);
 }
 
-static void handle_new_process( pid_t parent, pid_t child )
+void handle_new_process( pid_t parent_id, pid_t child_id )
 {
     // Copy the session information
-    dlog("%s: Registering "PID_F" with parent "PID_F"\n", __FUNCTION__, child, parent);
+    pid_state *child=&state[child_id]; // We actually want to create the state if it did not already exist
 
-    int_ptr process_type=state[parent].context_state[0];
+    if( child->state!=pid_state::INIT ) {
+        // Due to platform incompatibilities and other issues, we may be called several times over the same
+        // child. Don't make a fuss - just return.
 
-    if( (process_type&NEW_PROCESS_SAME_PARENT)==0 )
-        state[child].parent=parent;
-    else
-        state[child].parent=state[parent].parent;
+        dlog("%s: Process "PID_F" already registered - not performing any operation\n", __FUNCTION__, child_id );
 
-    state[child].session_id=state[parent].session_id;
-    state[state[child].parent].num_children++;
-
-    // if( (process_type&NEW_PROCESS_SAME_ROOT)==0 )
-        // XXX Need to contrast deep copy with shallow copy of root
-    state[child].root=state[parent].root;
-
-    // Whether the VM was copied or shared, the new process has the same static and shared memory
-    state[child].memory=state[parent].memory;
-    state[child].shared_memory=state[parent].shared_memory;
-    // If the VM is not shared, setting shared_memory but not shared_mem_local is an indication that the
-    // old memory needs to be freed
-    if( (process_type&NEW_PROCESS_SAME_VM)!=0 ) {
-        // The processes share the same VM - have them share the same shared memory
-        state[child].shared_mem_local=state[parent].shared_mem_local;
+        return;
     }
-}
 
-static pid_t first_child; // PID of first child
-static int comm_fd; // FD for communicating with the process "waiting" for the first child to exit
-static int num_processes; // Number of running processes
+    dlog("%s: Registering "PID_F" with parent "PID_F"\n", __FUNCTION__, child_id, parent_id );
+
+    // The platform may want to init the process in some way
+    ptlib_prepare(child_id);
+    child->state=pid_state::NONE;
+
+    pid_state *parent=lookup_state(parent_id);
+    if( parent!=NULL ) {
+        // If this assert fails, we somehow created a -1 process - not good
+        dlog(NULL);
+        assert(parent_id!=-1);
+
+        // This process is not a root process - it has a parent
+
+        int_ptr process_type=parent->context_state[0];
+
+        if( (process_type&NEW_PROCESS_SAME_PARENT)==0 )
+            child->parent=parent_id;
+        else
+            child->parent=parent->parent;
+
+        pid_state *child_parent=lookup_state(child->parent);
+        if( child_parent!=NULL ) {
+            child_parent->num_children++;
+        }
+
+        child->session_id=parent->session_id;
+
+        // if( (process_type&NEW_PROCESS_SAME_ROOT)==0 )
+        // XXX Need to contrast deep copy with shallow copy of root
+        child->root=parent->root;
+
+        // Whether the VM was copied or shared, the new process has the same static and shared memory
+        child->memory=parent->memory;
+        child->shared_memory=parent->shared_memory;
+        // If the VM is not shared, setting shared_memory but not shared_mem_local is an indication that the
+        // old memory needs to be freed
+        if( (process_type&NEW_PROCESS_SAME_VM)!=0 ) {
+            // The processes share the same VM - have them share the same shared memory
+            child->shared_mem_local=parent->shared_mem_local;
+        }
+
+        if( (process_type&NEW_PROCESS_SAME_DEBUGGER)!=0 ) {
+            // The process inherits the debugger from the parent
+            child->debugger=parent->debugger;
+        }
+
+        // Both parent and child need to call ptlib_fork_exit. We may need to copy the state
+        // from one to the other.
+        child->orig_sc=parent->orig_sc;
+        child->state=parent->state;
+        for( unsigned int i=0; i<PTLIB_STATE_SIZE; ++i )
+            child->saved_state[i]=parent->saved_state[i];
+
+        for( unsigned int i=0; i<=FORK_CONTEXT_SIZE; ++i )
+            child->context_state[i]=parent->context_state[i];
+    } else {
+        // This is a root process - no parent. Set it with the real session ID
+        child->session_id=getsid(child_id);
+    }
+
+    num_processes++;
+}
 
 int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long ret )
 {
     long sig=0;
 
-    //dlog("process_sigchld: "PID_F" state=%d, status=%x, ret=%d\n", pid, wait_state, status, ret );
+    pid_state *proc_state=lookup_state(pid);
+    if( wait_state!=NEWPROCESS && proc_state==NULL ) {
+        // The process does not exist!
+        // Register it
+        pid_t parent_pid=ptlib_get_parent(pid);
+        dlog("Caught unknown new process "PID_F", detected parent "PID_F"\n", pid, parent_pid );
+        dlog(NULL);
+        assert( parent_pid==0 || parent_pid==1 || state.find(parent_pid)!=state.end() ); // Make sure the parent is, indeed, ours
+
+        // Handle the process creation before handling the syscall return
+        process_sigchld( parent_pid, NEWPROCESS, status, pid );
+
+        // Handle the rest of the syscall as a return from a syscall
+        wait_state=SYSCALL;
+        proc_state=lookup_state(pid);
+        assert(proc_state!=NULL);
+        ret=proc_state->orig_sc;
+    }
 
     switch(wait_state) {
     case SYSCALL:
         {
             bool posttrap_always=false;
 
-            pid_state *proc_state=&state[pid];
             if( proc_state->state==pid_state::REDIRECT1 ) {
                 // REDIRECT1 is just a filler state between the previous call, where the arguments were set up and
                 // the call initiated, and the call's return (REDIRECT2). No need to actually call the handler
-                dlog(PID_F": Calling syscall %d redirected from %s\n", pid, ret, syscalls[proc_state->orig_sc].name );
+                dlog(PID_F": Calling syscall %ld redirected from %s\n", pid, ret, syscalls[proc_state->orig_sc].name );
                 proc_state->state=pid_state::REDIRECT2;
             } else if( proc_state->state==pid_state::REDIRECT2 || proc_state->state==pid_state::REDIRECT3 ) {
                 // REDIRECT2 means a return from a syscall generated by us.
                 // REDIRECT3 means entering a syscall generated by us, but for which the handler function would like
                 // to be notified (unlike REDIRECT1 above, which is short circuited)
                 if( proc_state->orig_sc!=SYS_execve ) {
-                    dlog(PID_F": Called syscall %d, redirected from %s\n", pid, ret, syscalls[proc_state->orig_sc].name );
+                    dlog(PID_F": Called syscall %ld, redirected from %s\n", pid, ret, syscalls[proc_state->orig_sc].name );
 
                     if( !syscalls[proc_state->orig_sc].func( ret, pid, proc_state ) )
                         sig=-1; // Mark for ptrace not to continue the process
                 } else {
                     // Special handling of the execve case
-                    dlog(PID_F": Called syscall %d, redirected from execve\n", pid, ret );
+                    dlog(PID_F": Called syscall %ld, redirected from execve\n", pid, ret );
 
                     if( !sys_execve( ret, pid, proc_state, posttrap_always ) )
                         sig=-1;
@@ -443,7 +555,7 @@ int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long
                 if( proc_state->state!=pid_state::ALLOCATE ) {
                     // Sanity check - returning from same syscall that got us in
                     if( proc_state->state==pid_state::RETURN && ret!=proc_state->orig_sc ) {
-                        dlog("process "PID_F" orig_sc=%d actual sc=%d state=%d\n", pid, proc_state->orig_sc, ret,
+                        dlog("process "PID_F" orig_sc=%d actual sc=%ld state=%s\n", pid, proc_state->orig_sc, ret,
                                 state2str(proc_state->state));
                         dlog(NULL);
                         assert( proc_state->state!=pid_state::RETURN || ret==proc_state->orig_sc );
@@ -528,7 +640,7 @@ int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long
         break;
     case SIGNAL:
         dlog(PID_F": Signal %s\n", pid, sig2str(ret));
-        if( state[pid].debugger==0 )
+        if( proc_state->debugger==0 )
             sig=ret;
         else {
             // Pass the signal to the debugger
@@ -537,8 +649,8 @@ int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long
             waiting.status()=status;
             getrusage( RUSAGE_CHILDREN, &waiting.usage() ); // XXX BUG this is the wrong function!
             waiting.debugonly()=true;
-            state[pid].trace_mode=TRACE_STOPPED2;
-            notify_parent( state[pid].debugger, waiting );
+            proc_state->trace_mode=TRACE_STOPPED2;
+            notify_parent( proc_state->debugger, waiting );
             sig=-1;
         }
         break;
@@ -554,10 +666,16 @@ int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long
             struct rusage rusage;
             getrusage( RUSAGE_CHILDREN, &rusage );
             handle_exit(pid, status, rusage );
-            if( pid==first_child && comm_fd!=-1 ) {
-                write( comm_fd, &status, sizeof(status) );
-                close( comm_fd );
-                comm_fd=-1;
+            
+            // If this was a root child, we may need to perform notification of exit status
+            MAP_CLASS<pid_t, int>::iterator root_child=root_children.find(pid);
+            if( root_child!=root_children.end() ) {
+                if( root_child->second!=-1 ) {
+                    write( root_child->second, &status, sizeof(status) );
+                    close( root_child->second );
+                }
+
+                root_children.erase(root_child);
             }
 
             num_processes--;
@@ -567,51 +685,184 @@ int process_sigchld( pid_t pid, enum PTLIB_WAIT_RET wait_state, int status, long
         {
             dlog(PID_F": Created new child process %ld\n", pid, ret);
             handle_new_process( pid, ret );
-            num_processes++;
         }
     }
 
     return sig;
 }
 
-int process_children(pid_t _first_child, int _comm_fd, pid_t session_id )
+bool attach_debugger( pid_t child, int socket )
 {
-    // Create a state for the first child
-    first_child=_first_child;
-    comm_fd=_comm_fd;
+    dlog(NULL);
 
+    // Attach a debugger to the child
+    if( ptrace(PTRACE_ATTACH, child, 0, 0)!=0 ) {
+        dlog("Could not start trace of process "PID_F": %s\n", child, strerror(errno) );
+
+        return false;
+    }
+    dlog("Debugger successfully attached to process "PID_F"\n", child );
+
+    // Let's free the process to do the exec
+    errno=0;
+    if( write( socket, "a", 1 )!=1 ) {
+        dlog("Couldn't free child process - write failed: %s\n", strerror(errno) );
+
+        return false;
+    }
+
+#if PTLIB_PARENT_CAN_WAIT
+    close( socket );
+    socket=-1;
+#endif
+
+    // If we start the processing loop too early, we might accidentally cath the "wait" where the master process (our grandparent)
+    // is waiting for our parent to terminate.
+    // In order to avoid that race, we wait until we notice that the process is sending itself a "USR1" signal to indicate it
+    // is ready.
+
+    bool sync=false;
+    while( !sync ) {
+        int status;
+
+        waitpid( child, &status, 0 );
+
+        if( WIFSTOPPED(status) ) {
+            switch( WSTOPSIG(status) ) {
+            case SIGUSR1:
+                // SIGUSR1 - that's our signal
+                dlog("Caught SIGUSR1 by child - start special handling\n");
+                ptrace( PTRACE_SYSCALL, child, 0, 0 );
+                sync=true;
+                break;
+            case SIGSTOP:
+                dlog("Caught SIGSTOP\n");
+                ptrace( PTRACE_CONT, child, 0, 0 ); // Continue the child in systrace mode
+                break;
+            case SIGTRAP:
+                dlog("Caught SIGTRAP\n");
+                ptrace( PTRACE_CONT, child, 0, 0 ); // Continue the child in systrace mode
+                break;
+            default:
+                dlog("Caught signal %d\n", WSTOPSIG(status) );
+                ptrace( PTRACE_CONT, child, 0, WSTOPSIG(status) );
+                break;
+            }
+        } else {
+            // Stopped for whatever other reason - just continue it
+            dlog("Another stop %x\n", status );
+            ptrace( PTRACE_CONT, child, 0, 0 );
+        }
+    }
+
+    // Child has started, and is debugged
+    root_children[child]=socket; // Mark this as a root child
+
+    handle_new_process( -1, child ); // No parent - a root process
+
+    return true;
+}
+
+// Do nothing signal handler for sigchld
+static void sigchld_handler(int signum)
+{
+}
+
+// Signify whether an alarm was received while we were waiting
+static bool alarm_happened=false;
+
+static void sigalrm_handler(int signum)
+{
+    alarm_happened=true;
+}
+
+int process_children( int master_socket )
+{
     // Initialize the ptlib library
     ptlib_init();
 
-    state[first_child]=pid_state();
-    state[first_child].session_id=session_id; // The initial session ID
     init_handlers();
     init_globals();
 
     dlog( "Begin the process loop\n" );
 
-    num_processes=1;
+    // Prepare the signal masks so we do not lose SIGCHLD while we wait
+
+    struct sigaction action;
+    memset( &action, 0, sizeof( action ) );
+
+    action.sa_handler=sigchld_handler;
+    sigemptyset( &action.sa_mask );
+    action.sa_flags=0;
+
+    sigaction( SIGCHLD, &action, NULL );
+
+    action.sa_handler=sigalrm_handler;
+    sigaction( SIGALRM, &action, NULL );
+
+    sigset_t orig_signals, child_signals;
+
+    sigemptyset( &child_signals );
+    sigaddset( &child_signals, SIGCHLD );
+    sigaddset( &child_signals, SIGALRM );
+    sigprocmask( SIG_BLOCK, &child_signals, &orig_signals );
+
+    sigdelset( &orig_signals, SIGCHLD );
+    sigdelset( &orig_signals, SIGALRM );
+
+    // Prepare the file descriptors
+    fd_set file_set;
+
+    FD_ZERO(&file_set);
+    if( master_socket>0 )
+        FD_SET(master_socket, &file_set);
 
     while(num_processes>0) {
         int status;
         pid_t pid;
         long ret;
         ptlib_extra_data data;
-        
+
         enum PTLIB_WAIT_RET wait_state;
-        if( !ptlib_wait( &pid, &status, &data ) ) {
-            dlog("ptlib_wait failed\n");
+        if( !ptlib_wait( &pid, &status, &data, true ) ) {
+            if( errno==EAGAIN ) {
+                // No process is waiting - halt until one exists or until the socket has something to say
+                fd_set read_set=file_set;
+                fd_set except_set=file_set;
+
+                if( pselect( master_socket+1, &read_set, NULL, &except_set, NULL, &orig_signals )>=0 ) {
+                    // Something happened on the socket - new root process?
+                    int session_socket=accept( master_socket, NULL, 0 );
+                    if( session_socket>=0 ) {
+                        pid_t child=-1;
+
+                        if( read( session_socket, &child, sizeof(child))>0 && child>0 ) {
+                            dlog("Got asynchronous request to attach to process "PID_F"\n", child);
+
+                            attach_debugger(child, session_socket);
+                        }
+                    }
+                }
+
+                // Did an alarm signal arrive?
+                if( alarm_happened ) {
+                    alarm_happened=false;
+
+                    dump_states();
+                }
+
+            } else if( errno==ECHILD ) {
+                // We should never get here. If we have no more children, we should have known about it already
+                dlog( "BUG - ptlib_wait failed with %s while numchildren is still %d\n", strerror(errno), num_processes );
+                dlog(NULL);
+                num_processes=0;
+            } else {
+                dlog("ptlib_wait failed %d: %s\n", errno, strerror(errno) );
+            }
+
             continue;
         }
 
-        // If this is the first time we see this process, we need to init the ptrace options for it
-        if( state[pid].state==pid_state::INIT ) {
-            dlog( PID_F": Init new process\n", pid);
-
-            ptlib_prepare(pid);
-            state[pid].state=pid_state::NONE;
-        }
-
         ret=ptlib_parse_wait( pid, status, &wait_state );
 
         long sig=process_sigchld( pid, wait_state, status, ret );
@@ -645,9 +896,13 @@ static bool allocate_shared_mem( pid_t pid, pid_state *state )
 {
     char filename[PATH_MAX];
 
-    const char *tmpdir=getenv("TMPDIR");
+    const char *tmpdir=getenv("FAKEROOT_TMPDIR");
+
+    if( tmpdir==NULL )
+        tmpdir=getenv("TMPDIR");
+
     if( tmpdir==NULL || strlen(tmpdir)>=PATH_MAX-sizeof("/fakeroot-ng.XXXXXX") )
-        tmpdir="/tmp";
+        tmpdir=DEFAULT_TMPDIR;
 
     sprintf(filename, "%s/fakeroot-ng.XXXXXX", tmpdir);
 
@@ -662,8 +917,7 @@ static bool allocate_shared_mem( pid_t pid, pid_state *state )
     }
 
     // Make sure that the file is big enough, but create it sparse
-    lseek( fd, shared_mem_size-1, SEEK_SET );
-    write( fd, filename, 1 );
+    ftruncate( fd, shared_mem_size );
 
     // Map the file into the local address space
     char *memory=(char *)mmap( NULL, shared_mem_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0 );
@@ -726,7 +980,8 @@ static bool handle_memory_allocation( int sc_num, pid_t pid, pid_state *state )
         // First step - mmap just returned
         if( ptlib_success( pid, sc_num ) ) {
             state->memory=(void *)ptlib_get_retval( pid );
-            dlog("handle_memory_allocation: "PID_F" allocated for our use %d bytes at %p\n", pid, static_mem_size, state->memory);
+            dlog("handle_memory_allocation: "PID_F" allocated for our use %lu bytes at %p\n", pid,
+                    (unsigned long)static_mem_size, state->memory);
 
             // "All" we need now is the shared memory. First, let's generate the local version for it.
             if(allocate_shared_mem( pid, state ))
@@ -778,7 +1033,7 @@ static bool handle_memory_allocation( int sc_num, pid_t pid, pid_state *state )
             ptlib_generate_syscall( pid, PREF_MMAP, (char *)state->memory+ptlib_prepare_memory_len() );
         } else {
             // open failed
-            dlog( "handle_memory_allocation: "PID_F" process failed to open %s: %s\n", pid, state->shared_mem_local.get(),
+            dlog( "handle_memory_allocation: "PID_F" process failed to open %s: %s\n", pid, state->shared_mem_local.getc(),
                 strerror(ptlib_get_error(pid, sc_num)) );
             ptlib_continue( PTRACE_KILL, pid, 0 );
             return false;
@@ -795,8 +1050,8 @@ static bool handle_memory_allocation( int sc_num, pid_t pid, pid_state *state )
         if( ptlib_success( pid, sc_num ) ) {
             // mmap succeeded
             state->shared_memory=(void *)(ptlib_get_retval( pid )+ptlib_prepare_memory_len());
-            dlog("handle_memory_allocation: "PID_F" allocated for our use %d shared bytes at %p\n", pid, shared_mem_size,
-                (char *)state->shared_memory-ptlib_prepare_memory_len());
+            dlog("handle_memory_allocation: "PID_F" allocated for our use %lu shared bytes at %p\n", pid,
+                    (unsigned long)shared_mem_size, (char *)state->shared_memory-ptlib_prepare_memory_len());
 
             // We now need to close the file descriptor
             ptlib_set_argument( pid, 1, state->context_state[1] );
@@ -882,3 +1137,31 @@ pid_state *lookup_state( pid_t pid ) {
 
     return NULL;
 }
+
+void delete_state( pid_t pid )
+{
+    pid_state *proc_state=lookup_state(pid);
+    assert(proc_state!=NULL);
+    assert(proc_state->state==pid_state::ZOMBIE);
+
+    if( (--proc_state->context_state[0])==0 )
+        state.erase(pid);
+}
+
+void dump_states()
+{
+    // Print the header
+    dlog("PID\tParent\tState\n");
+
+    for( map_class<pid_t, pid_state>::const_iterator i=state.begin(); i!=state.end(); ++i ) {
+        dlog(PID_F"\t"PID_F"\t%s", i->first, i->second.parent, state2str(i->second.state) );
+
+        if( i->second.state==pid_state::ZOMBIE ) {
+            dlog("(%d)", (int)i->second.context_state[0]);
+        }
+
+        dlog("\n");
+    }
+
+    dlog(NULL);
+}
